#pragma clang diagnostic push
#pragma ide diagnostic ignored "readability-make-member-function-const"
#pragma ide diagnostic ignored "readability-convert-member-functions-to-static"

#include <botan/base32.h>
#include <botan/base64.h>
#include <qrcodegen.hpp>
#include <png++/png.hpp>

#include "controllers/controllers.h"
#include "db/db.h"
#include "dto/dto.h"

std::string create_totp_qrcode(const db::User& user, const std::string& b32_secret) {
    const int qrcode_pixel_size = 4;

    std::stringstream code_ss;
    code_ss << "otpauth://totp/MFileserver:"
            << user.getValueOfName()
            << "?secret="
            << b32_secret
            << "&issuer=MFileserver";
    auto code = qrcodegen::QrCode::encodeText(code_ss.str().c_str(), qrcodegen::QrCode::Ecc::MEDIUM);
    const int mod_count = code.getSize();
    png::image<png::gray_pixel> image(mod_count*qrcode_pixel_size, mod_count*qrcode_pixel_size);
    for (int x = 0; x < mod_count; x++) for (int y = 0; y < mod_count; y++) {
        const bool mod = code.getModule(x, y);
        const int x_img_start = x * qrcode_pixel_size, y_img_start = y * qrcode_pixel_size;
        for (int x_img = x_img_start; x_img < x_img_start + qrcode_pixel_size; x_img++) for (int y_img = y_img_start; y_img < y_img_start + qrcode_pixel_size; y_img++)
            image[x_img][y_img] = mod ? 0 : 0xff;
    }
    std::stringstream image_ss;
    image.write_stream(image_ss);

    std::string image_str = image_ss.str();
    std::vector<uint8_t> secret(image_str.data(), image_str.data()+image_str.size());

    return "data:image/png;base64," + Botan::base64_encode(secret);
}

namespace api {
    void auth::tfa_setup(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);
        Json::Value &json = *req->jsonObject();
        try {
            bool mail = dto::json_get<bool>(json, "mail").value();

            auto secret_uchar = rng->random_vec(32);
            std::vector<char> secret(secret_uchar.data(), secret_uchar.data()+32);
            user.setTfaSecret(secret);

            db::MapperUser user_mapper(drogon::app().getDbClient());
            user_mapper.update(user);

            if (mail) {
                send_mail(user);
                cbk(dto::Responses::get_success_res());
            } else {
                std::string b32_secret = Botan::base32_encode(secret_uchar);
                b32_secret.erase(std::remove(b32_secret.begin(), b32_secret.end(), '='), b32_secret.end());
                std::string code = create_totp_qrcode(user, b32_secret);
                cbk(dto::Responses::get_tfa_setup_res(b32_secret, code));
            }
        } catch (const std::exception&) {
            cbk(dto::Responses::get_badreq_res("Validation error"));
        }
    }

    void auth::tfa_complete(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);
        Json::Value &json = *req->jsonObject();
        try {
            bool mail = dto::json_get<bool>(json, "mail").value();
            uint32_t code = std::stoi(dto::json_get<std::string>(json, "code").value());

            user.setTfaType(mail ? db::tfaTypes::EMAIL : db::tfaTypes::TOTP);

            if (!verify2fa(user, code))
                return cbk(dto::Responses::get_unauth_res("Incorrect 2fa"));

            db::MapperUser user_mapper(drogon::app().getDbClient());
            user_mapper.update(user);

            revoke_all(user);
            cbk(dto::Responses::get_success_res());
        } catch (const std::exception&) {
            cbk(dto::Responses::get_badreq_res("Validation error"));
        }
    }

    void auth::tfa_disable(req_type req, cbk_type cbk) {
        db::User user = dto::get_user(req);

        db::MapperUser user_mapper(drogon::app().getDbClient());
        user.setTfaType(db::tfaTypes::NONE);
        user_mapper.update(user);

        revoke_all(user);
        cbk(dto::Responses::get_success_res());
    }
}

#pragma clang diagnostic pop